import os
import emod_api.spatialreports.spatial as sr
import matplotlib.pyplot as plt
import numpy as np
from COMPS.Data import Experiment, Simulation, QueryCriteria
from COMPS import Client

"""
This script assumes local SpatialReport_XXX.bin files which have been downloaded (from COMPS)
using pyCOMPS getexpout function or equivalent. Note that this interacts with files on an experiment
basis, not simulation basis. It assumes the files are in a subdirectory named after the experiment id,
and then in subdirectories of that named after the simulation id. Note that it also assumes there is
a subdirectory in between to represent a chosen tag id, which defaults to "ref" if not used. I.e.
<exp_id>/
   ref/
      <sim1_id>/
          SpatialReport_XXX.bin
      <sim2_id>/
          SpatialReport_XXX.bin

The idea is that the data is most interesting not on a simulation basis, but for an experiment,
especially aggregated on a certain sweep param and value. This plot calculates means and plots those.
"""

def plot_spatial_prev( exp_id, chan="Prevalence", tag="ref" ):
    node_chan_data={}
    node_chan_means = {}

    if "=" not in tag:
        sims_path = os.path.join( exp_id, tag ) 
    else:
        sims_path = os.path.join( exp_id, "ref" )

    # If we are using tag info, we need to do a query, which shouldn't be the default.
    if tag != "ref":
        Client.login( "https://comps.idmod.org" )
        exp = Experiment.get( exp_id )
        sims = exp.get_simulations(QueryCriteria().select_children(['tags']))
    else:
        sims = os.listdir( sims_path )

    for sim in sims:
        if tag != "ref":
            tag_key = tag.split( "=" )[0]
            tag_value = tag.split( "=" )[1]
            print( f"tag_key={tag_key}, {sim.tags[ tag_key ]}, {tag_value}." )
            if tag_key in sim.tags and sim.tags[ tag_key ] != tag_value:
                continue
            sim_id = str(sim.id)
        else:
            sim_id = sim
        
        report_path = os.path.join( sims_path, sim_id, "SpatialReport_" + chan + ".bin" )
        data = sr.SpatialReport( report_path )
        for node_id in data.node_ids:
            chan_data = data[node_id].data
            if node_id not in node_chan_data:
                node_chan_data[ node_id ] = list()
                node_chan_means[ node_id ] = list()
            node_chan_data[ node_id ].append( chan_data )
            node_chan_means[ node_id ] = np.zeros(len(node_chan_data[node_id]))
    for node in node_chan_data.keys():
        node_chan_means[ node ] = np.mean( np.array( node_chan_data[ node ] ), axis=0 )
        plt.plot( node_chan_means[ node ], label=f"node={node}" )
    plt.xlabel( "Timestep" )
    plt.ylabel( chan )
    plt.title( f"Mean values of {chan} over time by node." )
    plt.legend()
    plt.show()


if __name__ == "__main__": 
    import argparse
    parser = argparse.ArgumentParser(description='Spatial Report Plotting')
    parser.add_argument('-c', '--channel', action='store', default="Prevalence", help='channel(s) to display [Prevalence]' )
    parser.add_argument('-e', '--experiment_id', action='store', default=None, help='experiment id to plot, data assumed to be local' ) 
    parser.add_argument('-t', '--tag', action='store', default="ref", help='tag constraint' ) 
    args = parser.parse_args()
    if not args.experiment_id:
        with open( "COMPS_ID", "r" ) as fp:
            args.experiment_id = fp.read() 
    if not args.channel:
        args.channel = 'Prevalence' # should not be necessary

    # check that folder with name experiment_id exists
    if not os.path.exists( str(args.experiment_id) ):
        raise ValueError( f"Don't see folder for {args.experiment_id}." )

    plot_spatial_prev( str(args.experiment_id), args.channel, tag=args.tag )
